
/*
 * bltUnixWindow.c --
 *
 * This module implements additional window functionality for the BLT
 * toolkit, such as transparent Tk windows, and reparenting Tk
 * windows.
 *
 *	Copyright 1993-2004 George A Howlett.
 *
 *	Permission is hereby granted, free of charge, to any person
 *	obtaining a copy of this software and associated documentation
 *	files (the "Software"), to deal in the Software without
 *	restriction, including without limitation the rights to use,
 *	copy, modify, merge, publish, distribute, sublicense, and/or
 *	sell copies of the Software, and to permit persons to whom the
 *	Software is furnished to do so, subject to the following
 *	conditions:
 *
 *	The above copyright notice and this permission notice shall be
 *	included in all copies or substantial portions of the
 *	Software.
 *
 *	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
 *	KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 *	WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 *	PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
 *	OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 *	OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 *	OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 *	SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <X11/Xlib.h>
#include <X11/Xproto.h>

#include <tkPort.h>
#include <tkInt.h>

#include "bltInt.h"

Window
Blt_GetParentWindow(Display *display, Window window)
{
    Window root, parent;
    Window *dummy;
    unsigned int count;

    if (XQueryTree(display, window, &root, &parent, &dummy, &count) > 0) {
	XFree(dummy);
	return parent;
    }
    return None;
}

/*
 *---------------------------------------------------------------------------
 *
 * Blt_GetWindowId --
 *
 *      Returns the XID for the Tk_Window given.  Starting in Tk 8.0,
 *      the toplevel widgets are wrapped by another window.  Currently
 *      there's no way to get at that window, other than what is done
 *      here: query the X window hierarchy and retrieve the parent.
 *
 * Results:
 *      Returns the X Window ID of the widget.  If it's a toplevel, then
 *	the XID of the wrapper is returned.
 *
 *---------------------------------------------------------------------------
 */
Window
Blt_GetWindowId(Tk_Window tkwin)
{
    Window window;

    Tk_MakeWindowExist(tkwin);
    window = Tk_WindowId(tkwin);
    if (Tk_IsTopLevel(tkwin)) {
	Window parent;

	parent = Blt_GetParentWindow(Tk_Display(tkwin), window);
	if (parent != None) {
	    window = parent;
	}
	window = parent;
    }
    return window;
}

/*
 *---------------------------------------------------------------------------
 *
 * XGeometryErrorProc --
 *
 *	Flags errors generated from XGetGeometry calls to the X server.
 *
 * Results:
 *	Always returns 0.
 *
 * Side Effects:
 *	Sets a flag, indicating an error occurred.
 *
 *---------------------------------------------------------------------------
 */
/* ARGSUSED */
static int
XGeometryErrorProc(ClientData clientData, XErrorEvent *errEventPtr)
{
    int *errorPtr = clientData;

    *errorPtr = FALSE;
    return 0;
}

int
Blt_GetWindowRegion(Display *display, Window window, int *xPtr, int *yPtr, 
		    int *widthPtr, int *heightPtr)
{
    Tk_ErrorHandler handler;
    Window root;
    int any = -1;
    int result;
    int x, y;
    unsigned int w, h, bw, depth;

    handler = Tk_CreateErrorHandler(display, any, X_GetGeometry, any, 
	XGeometryErrorProc, &result);
    result = XGetGeometry(display, window, &root, &x, &y, &w, &h, &bw, &depth);
    if (!result) {
	goto error;
    }
    if (widthPtr != NULL) {
	*widthPtr = (int)w;
    }
    if (heightPtr != NULL) {
	*heightPtr = (int)h;
    }
    if ((xPtr != NULL) || (yPtr != NULL)) {
	int rootX, rootY;
	rootX = rootY = 0;

	do {
	    Window *children, parent;
	    unsigned int n;
	    
	    parent = -1;
	    result = XGetGeometry(display, window, &root, &x, &y, &w, &h, 
				  &bw, &depth);
	    if (!result) {
		goto error;
	    }
	    rootX += x + bw;
	    rootY += y + bw;
	    result = XQueryTree(display, window, &root, &parent, &children, &n);
	    XFree(children);
	    if (!result) {
		goto error;
	    }
	    window = parent;
	} while (window != root);

	if (xPtr != NULL) {
	    *xPtr = rootX;
	}
	if (yPtr != NULL) {
	    *yPtr = rootY;
	}
    }
    Tk_DeleteErrorHandler(handler);
    XSync(display, False);
    return TCL_OK;
 error:
    Tk_DeleteErrorHandler(handler);
    XSync(display, False);
    fprintf(stderr, "failed to get window region\n");
    return TCL_ERROR;
}

int
Blt_GetWindowFromObj(Tcl_Interp *interp, Tcl_Obj *objPtr, Window *windowPtr)
{
    const char *string;

    string = Tcl_GetString(objPtr);
    if (string[0] == '.') {
	Tk_Window tkwin;

	tkwin = Tk_NameToWindow(interp, string, Tk_MainWindow(interp));
	if (tkwin == NULL) {
	    return TCL_ERROR;
	}
	if (Tk_WindowId(tkwin) == None) {
	    Tk_MakeWindowExist(tkwin);
	}
	*windowPtr = (Tk_IsTopLevel(tkwin)) ? Blt_GetWindowId(tkwin) : 
	    Tk_WindowId(tkwin);
    } else if (strcmp(string, "root") == 0) {
	*windowPtr = Tk_RootWindow(Tk_MainWindow(interp));
    } else {

	int id;

	if (Tcl_GetIntFromObj(interp, objPtr, &id) != TCL_OK) {
	    return TCL_ERROR;
	}
	*windowPtr = (Window)id;
    }
    return TCL_OK;
}

