
/*
 * bltUtil.c --
 *
 * This module implements utility procedures for the BLT toolkit.
 *
 *	Copyright 1991-2004 George A Howlett.
 *
 *	Permission is hereby granted, free of charge, to any person obtaining
 *	a copy of this software and associated documentation files (the
 *	"Software"), to deal in the Software without restriction, including
 *	without limitation the rights to use, copy, modify, merge, publish,
 *	distribute, sublicense, and/or sell copies of the Software, and to
 *	permit persons to whom the Software is furnished to do so, subject to
 *	the following conditions:
 *
 *	The above copyright notice and this permission notice shall be
 *	included in all copies or substantial portions of the Software.
 *
 *	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *	EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *	MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *	NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 *	LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 *	OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *	WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <stdarg.h>
#include <float.h>
#include <math.h>

#include "bltInt.h"
#include "bltOp.h"
#include "bltHash.h"
#include "bltDBuffer.h"

char* Blt_Strdup(const char *string)
{
    size_t size;
    char *ptr;

    size = strlen(string) + 1;
    ptr = malloc(size * sizeof(char));
    if (ptr != NULL) {
	strcpy(ptr, string);
    }
    return ptr;
}

void strtolower(char *s)
{
    while (*s != '\0') {
	*s = tolower((unsigned char)(*s));
	s++;
    }
}

int TclGetLong(
    Tcl_Interp *interp,		/* Interpreter used for error reporting if not
				 * NULL. */
    CONST char *src,		/* String containing a (possibly signed) long
				 * integer in a form acceptable to
				 * Tcl_GetLongFromObj(). */
    long *longPtr)		/* Place to store converted long result. */
{
    Tcl_Obj obj;
    int code;

    obj.refCount = 1;
    obj.bytes = (char *) src;
    obj.length = strlen(src);
    obj.typePtr = NULL;

    code = Tcl_GetLongFromObj(interp, &obj, longPtr);
    if (obj.refCount > 1) {
	Tcl_Panic("invalid sharing of Tcl_Obj on C stack");
    }
    return code;
}

/*
 *---------------------------------------------------------------------------
 *
 * Blt_GetDoubleFromString --
 *
 *      Converts a string into a double precision number.  This differs from
 *	Tcl's version in that it also allows NaN and +/-Inf.  There are 
 *	cases where NaNs are used to indicate holes in the data. 
 *
 * Results:
 *      Returns a standard Tcl result.
 *
 *---------------------------------------------------------------------------
 */
int
Blt_GetDoubleFromString(Tcl_Interp *interp, const char *s, double *valuePtr)
{
    char *end;
    double d;
    
    errno = 0;
    d = strtod(s, &end); /* INTL: Tcl source. */
    if (end == s) {
	badDouble:
        if (interp != (Tcl_Interp *) NULL) {
            Tcl_AppendResult(interp,
		"expected floating-point number but got \"", s, "\"", 
		(char *) NULL);
        }
	return TCL_ERROR;
    }
    if (errno != 0 && (d == HUGE_VAL || d == -HUGE_VAL || d == 0)) {
        if (interp != (Tcl_Interp *) NULL) {
	    char msg[64 + TCL_INTEGER_SPACE];
	
	    sprintf(msg, "unknown floating-point error, errno = %d", errno);
	    Tcl_AppendToObj(Tcl_GetObjResult(interp), msg, -1);
	    Tcl_SetErrorCode(interp, "ARITH", "UNKNOWN", msg, (char *) NULL);
        }
	return TCL_ERROR;
    }
    while ((*end != 0) && isspace((unsigned char)(*end))) { /* INTL: ISO space. */
	end++;
    }
    if (*end != 0) {
	goto badDouble;
    }
    *valuePtr = d;
    return TCL_OK;
}

/*
 *---------------------------------------------------------------------------
 *
 * Blt_GetDoubleFromObj --
 *
 *      Converts a Tcl_Obj into a double precision number.  This differs from
 *	Tcl's version in that it also allows NaN and +/-Inf.  There are 
 *	cases where NaNs are used to indicate holes in the data. 
 *
 * Results:
 *      Returns a standard Tcl result.
 *
 * Side Effects:
 *	tclDoubleType is no longer available (in 8.5) as a global variable.
 *	We have to get a double obj and save its type pointer.
 *
 *---------------------------------------------------------------------------
 */
int
Blt_GetDoubleFromObj(Tcl_Interp *interp, Tcl_Obj *objPtr, double *valuePtr)
{
    const char *string;
    static Tcl_ObjType *tclDoubleTypePtr = NULL;

    if (tclDoubleTypePtr == NULL) {
	Tcl_Obj *objPtr;
	
	objPtr = Tcl_NewDoubleObj(0.0);
	tclDoubleTypePtr = (Tcl_ObjType*)(objPtr->typePtr);
	Tcl_DecrRefCount(objPtr);
    }
    if (objPtr->typePtr == tclDoubleTypePtr) {
	*valuePtr = objPtr->internalRep.doubleValue;
	return TCL_OK;
    }
    string = Tcl_GetString(objPtr);
    return Blt_GetDoubleFromString(interp, string, valuePtr);
}

/*
 *---------------------------------------------------------------------------
 *
 * Blt_CompareDictionary
 *
 *	This function compares two strings as if they were being used in
 *	an index or card catalog.  The case of alphabetic characters is
 *	ignored, except to break ties.  Thus "B" comes before "b" but
 *	after "a".  Also, integers embedded in the strings compare in
 *	numerical order.  In other words, "x10y" comes after "x9y", not
 *      before it as it would when using strcmp().
 *
 * Results:
 *      A negative result means that the first element comes before the
 *      second, and a positive result means that the second element
 *      should come first.  A result of zero means the two elements
 *      are equal and it doesn't matter which comes first.
 *
 * Side effects:
 *	None.
 *
 *---------------------------------------------------------------------------
 */

#if HAVE_UTF
int
Blt_DictionaryCompare(const char *left, const char *right)
{
    Tcl_UniChar uniLeft, uniRight, uniLeftLower, uniRightLower;
    int diff, zeros;
    int secondaryDiff = 0;

    for(;;) {
	if ((isdigit((unsigned char)(*right))) && (isdigit((unsigned char)(*left)))) { 
	    /*
	     * There are decimal numbers embedded in the two strings.  Compare
	     * them as numbers, rather than strings.  If one number has more
	     * leading zeros than the other, the number with more leading
	     * zeros sorts later, but only as a secondary choice.
	     */

	    zeros = 0;
	    while ((*right == '0') && (isdigit((unsigned char)(right[1])))) {
		right++;
		zeros--;
	    }
	    while ((*left == '0') && (isdigit((unsigned char)(left[1])))) {
		left++;
		zeros++;
	    }
	    if (secondaryDiff == 0) {
		secondaryDiff = zeros;
	    }

	    /*
	     * The code below compares the numbers in the two strings without
	     * ever converting them to integers.  It does this by first
	     * comparing the lengths of the numbers and then comparing the
	     * digit values.
	     */

	    diff = 0;
	    for (;;) {
		if (diff == 0) {
		    diff = (unsigned char)(*left) - (unsigned char)(*right);
		}
		right++;
		left++;

		/* Ignore commas in numbers. */
		if (*left == ',') {
		    left++;
		}
		if (*right == ',') {
		    right++;
		}

		if (!isdigit((unsigned char)(*right))) { /* INTL: digit */
		    if (isdigit((unsigned char)(*left))) { /* INTL: digit */
			return 1;
		    } else {
			/*
			 * The two numbers have the same length. See
			 * if their values are different.
			 */

			if (diff != 0) {
			    return diff;
			}
			break;
		    }
		} else if (!isdigit((unsigned char)(*left))) { /* INTL: digit */
		    return -1;
		}
	    }
	    continue;
	}

	/*
	 * Convert character to Unicode for comparison purposes.  If either
	 * string is at the terminating null, do a byte-wise comparison and
	 * bail out immediately.
	 */
	if ((*left != '\0') && (*right != '\0')) {
	    left += Tcl_UtfToUniChar(left, &uniLeft);
	    right += Tcl_UtfToUniChar(right, &uniRight);
	    /*
	     * Convert both chars to lower for the comparison, because
	     * dictionary sorts are case insensitve.  Convert to lower, not
	     * upper, so chars between Z and a will sort before A (where most
	     * other interesting punctuations occur)
	     */
	    uniLeftLower = Tcl_UniCharToLower(uniLeft);
	    uniRightLower = Tcl_UniCharToLower(uniRight);
	} else {
	    diff = (unsigned char)(*left) - (unsigned char)(*right);
	    break;
	}

        diff = uniLeftLower - uniRightLower;
        if (diff) {
	    return diff;
	} else if (secondaryDiff == 0) {
	    if (Tcl_UniCharIsUpper(uniLeft) &&
		    Tcl_UniCharIsLower(uniRight)) {
		secondaryDiff = -1;
	    } else if (Tcl_UniCharIsUpper(uniRight)
		    && Tcl_UniCharIsLower(uniLeft)) {
		secondaryDiff = 1;
	    }
        }
    }
    if (diff == 0) {
	diff = secondaryDiff;
    }
    return diff;
}

#else 

int
Blt_DictionaryCompare(const char *left, const char *right) 
{
    int diff, zeros;
    int secondaryDiff = 0;

    while (1) {
	if (isdigit((unsigned char)(*right)) && isdigit((unsigned char)(*left))) {
	    /*
	     * There are decimal numbers embedded in the two strings.  Compare
	     * them as numbers, rather than strings.  If one number has more
	     * leading zeros than the other, the number with more leading
	     * zeros sorts later, but only as a secondary choice.
	     */

	    zeros = 0;
	    while ((*right == '0') && (isdigit((unsigned char)(right[1])))) {
		right++;
		zeros--;
	    }
	    while ((*left == '0') && (isdigit((unsigned char)(left[1])))) {
		left++;
		zeros++;
	    }
	    if (secondaryDiff == 0) {
		secondaryDiff = zeros;
	    }

	    /*
	     * The code below compares the numbers in the two strings without
	     * ever converting them to integers.  It does this by first
	     * comparing the lengths of the numbers and then comparing the
	     * digit values.
	     */

	    diff = 0;
	    while (1) {
		if (diff == 0) {
		    diff = (unsigned char)(*left) - (unsigned char)(*right);
		}
		right++;
		left++;
		/* Ignore commas in numbers. */
		if (*left == ',') {
		    left++;
		}
		if (*right == ',') {
		    right++;
		}
		if (!isdigit((unsigned char)(*right))) {
		    if (isdigit((unsigned char)(*left))) {
			return 1;
		    } else {
			/*
			 * The two numbers have the same length. See
			 * if their values are different.
			 */

			if (diff != 0) {
			    return diff;
			}
			break;
		    }
		} else if (!isdigit((unsigned char)(*left))) {
		    return -1;
		}
	    }
	    continue;
	}
        diff = (unsigned char)(*left) - (unsigned char)(*right);
        if (diff) {
            if (isupper((unsigned char)(*left)) && islower((unsigned char)(*right))) {
                diff = (unsigned char)(tolower(*left)) - (unsigned char)(*right);
                if (diff) {
		    return diff;
                } else if (secondaryDiff == 0) {
		    secondaryDiff = -1;
                }
            } else if (isupper((unsigned char)(*right)) && islower((unsigned char)(*left))) {
                diff = (unsigned char)(*left) - (unsigned char)(tolower((unsigned char)(*right)));
                if (diff) {
		    return diff;
                } else if (secondaryDiff == 0) {
		    secondaryDiff = 1;
                }
            } else {
                return diff;
            }
        }
        if (*left == 0) {
	    break;
	}
        left++;
        right++;
    }
    if (diff == 0) {
	diff = secondaryDiff;
    }
    return diff;
}
#endif

/*ARGSUSED*/
void
Blt_Panic TCL_VARARGS_DEF(const char *, arg1)
{
    const char *format;
    va_list args;

    format = TCL_VARARGS_START(const char *, arg1, args);
    vfprintf(stderr, format, args);
    fprintf(stderr, "\n");
    fflush(stderr);
    abort();
}

/*ARGSUSED*/
int sprintf_s(char *s, size_t size, const char *fmt, /*args*/ ...) 
{
    va_list ap;
    int n;

    va_start(ap, fmt);
    n = vsnprintf(s, size, fmt, ap);
    if ((n != (int)size) && (size > 0)) {
	s[size-1] = '\0';
    }
    va_end(ap);
    return n;
}

static char stringRep[200];

const char *
Blt_Itoa(int value)
{
    sprintf_s(stringRep, 200, "%d", value);
    return stringRep;
}

const char *
Blt_Ltoa(long value)
{
    sprintf_s(stringRep, 200, "%ld", value);
    return stringRep;
}

const char *
Blt_Utoa(unsigned int value)
{
    sprintf_s(stringRep, 200, "%u", value);
    return stringRep;
}

const char *
Blt_Dtoa(Tcl_Interp *interp, double value)
{
    Tcl_PrintDouble(interp, value, stringRep);
    return stringRep;
}

int
Blt_GlobalEvalObjv(Tcl_Interp *interp, int objc, Tcl_Obj *const *objv)
{
    int i;
    int result;

    for (i = 0; i < objc; i++) {
	Tcl_IncrRefCount(objv[i]);
    }
    result = Tcl_EvalObjv(interp, objc, objv, TCL_EVAL_GLOBAL);
    for (i = 0; i < objc; i++) {
	Tcl_DecrRefCount(objv[i]);
    }
    return result;
}

int
Blt_GetCountFromObj(
    Tcl_Interp *interp,
    Tcl_Obj *objPtr,
    int check,				/* Can be COUNT_POS, COUNT_NNEG,
					 * or COUNT_ANY, */
    long *valuePtr)
{
    long count;

    if (Tcl_GetLongFromObj(interp, objPtr, &count) != TCL_OK) {
	return TCL_ERROR;
    }
    switch (check) {
    case COUNT_NNEG:
	if (count < 0) {
	    Tcl_AppendResult(interp, "bad value \"", Tcl_GetString(objPtr), 
		"\": can't be negative", (char *)NULL);
	    return TCL_ERROR;
	}
	break;
    case COUNT_POS:
	if (count <= 0) {
	    Tcl_AppendResult(interp, "bad value \"", Tcl_GetString(objPtr), 
		"\": must be positive", (char *)NULL);
	    return TCL_ERROR;
	}
	break;
    case COUNT_ANY:
	break;
    }
    *valuePtr = count;
    return TCL_OK;
}

/*
 *---------------------------------------------------------------------------
 *
 * BinaryOpSearch --
 *
 *      Performs a binary search on the array of command operation
 *      specifications to find a partial, anchored match for the given
 *      operation string.
 *
 * Results:
 *	If the string matches unambiguously the index of the specification in
 *	the array is returned.  If the string does not match, even as an
 *	abbreviation, any operation, -1 is returned.  If the string matches,
 *	but ambiguously -2 is returned.
 *
 *---------------------------------------------------------------------------
 */
static int
BinaryOpSearch(Blt_OpSpec *specs, int nSpecs, const char *string, int length)
{
    char c;
    int high, low;

    low = 0;
    high = nSpecs - 1;
    c = string[0];
    while (low <= high) {
	Blt_OpSpec *specPtr;
	int compare;
	int median;
	
	median = (low + high) >> 1;
	specPtr = specs + median;

	/* Test the first character */
	compare = c - specPtr->name[0];
	if (compare == 0) {
	    /* Now test the entire string */
	    compare = strncmp(string, specPtr->name, length);
	    if (compare == 0) {
		if ((int)length < specPtr->minChars) {
		    return -2;	/* Ambiguous operation name */
		}
	    }
	}
	if (compare < 0) {
	    high = median - 1;
	} else if (compare > 0) {
	    low = median + 1;
	} else {
	    return median;		/* Op found. */
	}
    }
    return -1;				/* Can't find operation */
}


/*
 *---------------------------------------------------------------------------
 *
 * LinearOpSearch --
 *
 *      Performs a binary search on the array of command operation
 *      specifications to find a partial, anchored match for the given
 *      operation string.
 *
 * Results:
 *	If the string matches unambiguously the index of the specification in
 *	the array is returned.  If the string does not match, even as an
 *	abbreviation, any operation, -1 is returned.  If the string matches,
 *	but ambiguously -2 is returned.
 *
 *---------------------------------------------------------------------------
 */
static int
LinearOpSearch(Blt_OpSpec *specs, int nSpecs, const char *string, int length)
{
    Blt_OpSpec *specPtr;
    char c;
    int nMatches, last;
    int i;

    c = string[0];
    nMatches = 0;
    last = -1;
    for (specPtr = specs, i = 0; i < nSpecs; i++, specPtr++) {
	if ((c == specPtr->name[0]) && 
	    (strncmp(string, specPtr->name, length) == 0)) {
	    last = i;
	    nMatches++;
	    if ((int)length == specPtr->minChars) {
		break;
	    }
	}
    }
    if (nMatches > 1) {
	return -2;			/* Ambiguous operation name */
    } 
    if (nMatches == 0) {
	return -1;			/* Can't find operation */
    } 
    return last;			/* Op found. */
}

/*
 *---------------------------------------------------------------------------
 *
 * Blt_GetOpFromObj --
 *
 *      Find the command operation given a string name.  This is useful where
 *      a group of command operations have the same argument signature.
 *
 * Results:
 *      If found, a pointer to the procedure (function pointer) is returned.
 *      Otherwise NULL is returned and an error message containing a list of
 *      the possible commands is returned in interp->result.
 *
 *---------------------------------------------------------------------------
 */
void *
Blt_GetOpFromObj(
    Tcl_Interp *interp,			/* Interpreter to report errors to */
    int nSpecs,				/* # of specifications in array */
    Blt_OpSpec *specs,			/* Op specification array */
    int operPos,			/* Position of operation in argument
					 * list. */
    int objc,				/* # of arguments in the argument *
					 * vector.  This includes any
					 * prefixed arguments */
    Tcl_Obj *const *objv,		/* Argument vector */
    int flags)
{
    Blt_OpSpec *specPtr;
    const char *string;
    int length;
    int n;

    if (objc <= operPos) {		/* No operation argument */
	Tcl_AppendResult(interp, "wrong # args: ", (char *)NULL);
      usage:
	Tcl_AppendResult(interp, "should be one of...", (char *)NULL);
	for (n = 0; n < nSpecs; n++) {
	    int i;

	    Tcl_AppendResult(interp, "\n  ", (char *)NULL);
	    for (i = 0; i < operPos; i++) {
		Tcl_AppendResult(interp, Tcl_GetString(objv[i]), " ", 
			 (char *)NULL);
	    }
	    specPtr = specs + n;
	    Tcl_AppendResult(interp, specPtr->name, " ", specPtr->usage,
		(char *)NULL);
	}
	return NULL;
    }
    string = Tcl_GetStringFromObj(objv[operPos], &length);
    if (flags & BLT_OP_LINEAR_SEARCH) {
	n = LinearOpSearch(specs, nSpecs, string, length);
    } else {
	n = BinaryOpSearch(specs, nSpecs, string, length);
    }
    if (n == -2) {
	char c;

	Tcl_AppendResult(interp, "ambiguous", (char *)NULL);
	if (operPos > 2) {
	    Tcl_AppendResult(interp, " ", Tcl_GetString(objv[operPos - 1]), 
		(char *)NULL);
	}
	Tcl_AppendResult(interp, " operation \"", string, "\" matches: ",
	    (char *)NULL);

	c = string[0];
	for (n = 0; n < nSpecs; n++) {
	    specPtr = specs + n;
	    if ((c == specPtr->name[0]) &&
		(strncmp(string, specPtr->name, length) == 0)) {
		Tcl_AppendResult(interp, " ", specPtr->name, (char *)NULL);
	    }
	}
	return NULL;

    } else if (n == -1) {	      /* Can't find operation, display help */
	Tcl_AppendResult(interp, "bad", (char *)NULL);
	if (operPos > 2) {
	    Tcl_AppendResult(interp, " ", Tcl_GetString(objv[operPos - 1]), 
		(char *)NULL);
	}
	Tcl_AppendResult(interp, " operation \"", string, "\": ", (char *)NULL);
	goto usage;
    }
    specPtr = specs + n;
    if ((objc < specPtr->minArgs) || 
	((specPtr->maxArgs > 0) && (objc > specPtr->maxArgs))) {
	int i;

	Tcl_AppendResult(interp, "wrong # args: should be \"", (char *)NULL);
	for (i = 0; i < operPos; i++) {
	    Tcl_AppendResult(interp, Tcl_GetString(objv[i]), " ", 
		(char *)NULL);
	}
	Tcl_AppendResult(interp, specPtr->name, " ", specPtr->usage, "\"",
	    (char *)NULL);
	return NULL;
    }
    return specPtr->proc;
}

