/*
 * bltWindow.c --
 *
 * This module implements additional window functions for the BLT
 * toolkit.
 *
 *	Copyright 1991-2004 George A Howlett.
 *
 *	Permission is hereby granted, free of charge, to any person
 *	obtaining a copy of this software and associated documentation
 *	files (the "Software"), to deal in the Software without
 *	restriction, including without limitation the rights to use,
 *	copy, modify, merge, publish, distribute, sublicense, and/or
 *	sell copies of the Software, and to permit persons to whom the
 *	Software is furnished to do so, subject to the following
 *	conditions:
 *
 *	The above copyright notice and this permission notice shall be
 *	included in all copies or substantial portions of the
 *	Software.
 *
 *	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
 *	KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 *	WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 *	PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
 *	OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 *	OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 *	OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 *	SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <X11/Xlib.h>
#include <X11/Xproto.h>

#include <tkPort.h>
#include <tkInt.h>

#include "bltInt.h"
#include "bltWindow.h"
#include "bltHash.h"

/* 
 * Cache drawable information.
 *
 * There are specific times when we need information about the drawable
 * that is not available via the normal argument passing.  For example,
 * XFT font rendering requires that the correct depth XftDraw be used.
 * If we create a painter from a drawable, we need to know its depth,
 * colormap, and visual also.
 * 
 * We can usually get this information using XGetWindowAttributes at
 * the cost of a round-trip to the X server.  Instead we'll cache it
 * in a hash table.  We don't have to cache every window. We need to
 * watch destroy events and freeing pixmaps.
 */

typedef struct {
  Display *display;
  Drawable drawable;
} DrawableKey;

static Blt_HashTable attribTable;
static int initialized = FALSE;

Blt_DrawableAttributes* Blt_GetDrawableAttribs(Display *display, 
					       Drawable drawable)
{
  if (drawable != None) {
    Blt_HashEntry *hPtr;
    DrawableKey key;

    if (!initialized) {
      Blt_InitHashTable(&attribTable, sizeof(DrawableKey)/sizeof(int));
      initialized = TRUE;
    }
    memset(&key, 0, sizeof(key));
    key.drawable = drawable;
    key.display = display;
    hPtr = Blt_FindHashEntry(&attribTable, &key);
    if (hPtr != NULL) {
      return Blt_GetHashValue(hPtr);
    }
  }
  return NULL;		/* Don't have any information about this
			 * drawable. */
}

void Blt_SetDrawableAttribs(Display *display, Drawable drawable, int depth,
			    int width, int height, Colormap colormap,
			    Visual *visual)
{
  if (drawable != None) {
    Blt_DrawableAttributes *attrPtr;
    Blt_HashEntry *hPtr;
    int isNew;
    DrawableKey key;

    if (!initialized) {
      Blt_InitHashTable(&attribTable, sizeof(DrawableKey)/sizeof(int));
      initialized = TRUE;
    }
    memset(&key, 0, sizeof(key));
    key.drawable = drawable;
    key.display = display;
    hPtr = Blt_CreateHashEntry(&attribTable, &key, &isNew);
    if (isNew) {
      attrPtr = malloc(sizeof(Blt_DrawableAttributes));
      Blt_SetHashValue(hPtr, attrPtr);
    }  else {
      attrPtr = Blt_GetHashValue(hPtr);
    }
    /* Set or reset information for drawable. */
    attrPtr->id = drawable;
    attrPtr->depth = depth;
    attrPtr->colormap = colormap;
    attrPtr->visual = visual;
    attrPtr->width = width;
    attrPtr->height = height;
  }
}

/*
 *---------------------------------------------------------------------------
 *
 * Blt_FindChild --
 *
 *      Performs a linear search for the named child window in a given
 *	parent window.
 *
 *	This can be done via Tcl, but not through Tk's C API.  It's 
 *	simple enough, if you peek into the Tk_Window structure.
 *
 * Results:
 *      The child Tk_Window. If the named child can't be found, NULL
 *	is returned.
 *
 *---------------------------------------------------------------------------
 */

Tk_Window Blt_FindChild(Tk_Window parent, char *name)
{
  TkWindow *winPtr;
  TkWindow *parentPtr = (TkWindow *)parent;

  for (winPtr = parentPtr->childList; winPtr != NULL; 
       winPtr = winPtr->nextPtr) {
    if (strcmp(name, winPtr->nameUid) == 0) {
      return (Tk_Window)winPtr;
    }
  }
  return NULL;
}

/*
 *---------------------------------------------------------------------------
 *
 * Blt_FirstChildWindow --
 *
 *      Performs a linear search for the named child window in a given
 *	parent window.
 *
 *	This can be done via Tcl, but not through Tk's C API.  It's 
 *	simple enough, if you peek into the Tk_Window structure.
 *
 * Results:
 *      The child Tk_Window. If the named child can't be found, NULL
 *	is returned.
 *
 *---------------------------------------------------------------------------
 */

Tk_Window Blt_FirstChild(Tk_Window parent)
{
  TkWindow *parentPtr = (TkWindow *)parent;
  return (Tk_Window)parentPtr->childList;
}

/*
 *---------------------------------------------------------------------------
 *
 * Blt_FindChild --
 *
 *      Performs a linear search for the named child window in a given
 *	parent window.
 *
 *	This can be done via Tcl, but not through Tk's C API.  It's 
 *	simple enough, if you peek into the Tk_Window structure.
 *
 * Results:
 *      The child Tk_Window. If the named child can't be found, NULL
 *	is returned.
 *
 *---------------------------------------------------------------------------
 */

Tk_Window Blt_NextChild(Tk_Window tkwin)
{
  TkWindow *winPtr = (TkWindow *)tkwin;

  if (winPtr == NULL) {
    return NULL;
  }
  return (Tk_Window)winPtr->nextPtr;
}

/*
 *---------------------------------------------------------------------------
 *
 * Blt_Toplevel --
 *
 *      Climbs up the widget hierarchy to find the top level window of
 *      the window given.
 *
 * Results:
 *      Returns the Tk_Window of the toplevel widget.
 *
 *---------------------------------------------------------------------------
 */

Tk_Window Blt_Toplevel(Tk_Window tkwin)
{
  while (!Tk_IsTopLevel(tkwin)) {
    tkwin = Tk_Parent(tkwin);
  }
  return tkwin;
}

void Blt_SetWindowInstanceData(Tk_Window tkwin, ClientData instanceData)
{
  TkWindow *winPtr = (TkWindow *)tkwin;

  winPtr->instanceData = instanceData;
}

ClientData Blt_GetWindowInstanceData(Tk_Window tkwin)
{
  TkWindow *winPtr;

  while (tkwin != NULL) {
    winPtr = (TkWindow *)tkwin;
    if (winPtr->instanceData != NULL) {
      return (ClientData)winPtr->instanceData;
    }
    tkwin = Tk_Parent(tkwin);
  }
  return NULL;
}
